import org.graphstream.algorithm.Dijkstra;
import org.graphstream.algorithm.Dijkstra.Element;
import org.graphstream.algorithm.generator.Generator;
import org.graphstream.algorithm.generator.RandomGenerator;
import org.graphstream.graph.Edge;
import org.graphstream.graph.Graph;
import org.graphstream.graph.Node;
import org.graphstream.graph.implementations.SingleGraph;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.TimeUnit;

import static org.graphstream.algorithm.Toolkit.randomNode;

public class Djikstra {

    //this methode allow us to generate a random graph (its size is taille)
    private static Graph generateRandomGraph(int taille, int averageDegre) {
        Graph graph2 = new SingleGraph("Random");
        Generator gen = new RandomGenerator(averageDegre);
        gen.addSink(graph2);
        gen.begin();
        for (int i = 0; i < taille; i++)
            gen.nextEvents();
        gen.end();

        //add a random poids to each edges
        Random rand = new Random();
        for (Edge e : graph2.getEachEdge()) {
            int a = rand.nextInt(40) + 1;
            e.addAttribute("poids", a);
        }

        //add parent(null) and distance(infinity) attributes
        for (Node n : graph2.getEachNode()) {
            n.addAttribute("distance", Integer.MAX_VALUE);
            n.addAttribute("parent", 0);
        }

        return graph2;
    }

    /*the main function which allow us to calculate the distance
        from the node passed in parameters to each node
     */
    public static void dijkstra(Graph g, Node depart) {

        //create a Linkedlist which will contain the nodes
        LinkedList<Node> elementsTries = new LinkedList();

        //change depart distance to 0 then add it to the LinkedList
        depart.setAttribute("distance", 0);
        elementsTries.add(depart);

        //to iterate on all the neighbours
        Iterator<Node> neighbours;

        while (!elementsTries.isEmpty()) {

            //get the node with min distance in the lifo
            int minInt = Integer.MAX_VALUE;
            Node u = depart;
            for (int i = 0; i < elementsTries.size(); i++) {
                if ((int) elementsTries.get(i).getAttribute("distance") < minInt) {
                    minInt = (int) elementsTries.get(i).getAttribute("distance");
                    u = elementsTries.get(i);
                }
                elementsTries.remove(u);
            }

            neighbours = u.getNeighborNodeIterator();
            //iterate on all the neighbours
            while (neighbours.hasNext()) {
                Node neighbour = neighbours.next();

                /*perform exactly the same calculations as the studied algorithme
                on the course
                 */
                if (u.hasEdgeToward(neighbour)) {
                    int Wuv = u.getEdgeBetween(neighbour).getAttribute("poids");

                    if (((int) u.getAttribute("distance")) + Wuv < (int) neighbour.getAttribute("distance")) {
                        neighbour.setAttribute("distance", (int) u.getAttribute("distance") + Wuv);
                        neighbour.setAttribute("parent", u);
                        elementsTries.add(neighbour);
                    }
                }
            }
        }
    }


    //garaphstream Djikstra
    public static void graphStramDjikstra(Graph graph, Node depart) {
        Dijkstra dijkstra = new Dijkstra(Element.EDGE, "result", "length");

        // Compute the shortest paths in g from depart to all nodes
        dijkstra.init(graph);
        dijkstra.setSource(depart);
        dijkstra.compute();
    }

    //create file to draw the curves( we will use it inside the methode below )
    private static void createFile(String fileName, long time, int nodeNumber, int averagedegre) {
        try {
            String filepath = System.getProperty("user.dir") + "/src/main/resources/averagedegre" + averagedegre + File.separator + fileName;

            //create file or append
            FileWriter fw = new FileWriter(filepath, true);
            BufferedWriter bw = new BufferedWriter(fw);
            bw.write(nodeNumber + "     " + time + "\n");
            bw.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /*1-calculate the execution time  of the graph stream djikstra's algo
      2- create a file with two columns (first ==> number of nodes, second ==>execution time)
      3-do the same thing with my algo
     */
    public static void printTime(int taille, int averageDegre) {
        Graph g = generateRandomGraph(taille, averageDegre);
        Node n = randomNode(g);
        System.out.println();
        long startTime = System.nanoTime();

        dijkstra(g, n);

        long endTime = System.nanoTime();
        long totalTime = endTime - startTime;
        long convert = TimeUnit.MILLISECONDS.convert(totalTime, TimeUnit.NANOSECONDS);
        System.out.println("Le temps d'execution avec mon algorithme est :" + convert + " ms");
        //create the file
        createFile("monAlgo(" + averageDegre + ")txt", convert, taille, averageDegre);


        long startTim = System.nanoTime();

        graphStramDjikstra(g, n);

        long endTim = System.nanoTime();
        long totalTim = endTim - startTim;
        long conver = TimeUnit.MILLISECONDS.convert(totalTim, TimeUnit.NANOSECONDS);
        System.out.println("Le temps d'execution avec djikstra de graphStream :" + conver + " ms");
        createFile("djikstra(" + averageDegre + ")txt", conver, taille, averageDegre);
        System.out.println();

    }

    public static void main(String[] args) throws IOException {

        //test many graph with averageDegre=3
        int i = 0;
        while (i <= 100000) {
            System.out.println("Test avec un graphe de taille " + i);
            printTime(i, 3);
            System.out.println("*************************************************");
            i += 10000;
        }


        //test many graph with averageDegre=10
        i = 0;
        while (i <= 80000) {
            System.out.println("Test avec un graphe de taille " + i);
            printTime(i, 10);
            System.out.println("*************************************************");
            i += 8000;
        }

        //test many graph with averageDegre=50
        i = 0;
        while (i <= 10000) {
            System.out.println("Test avec un graphe de taille " + i);
            printTime(i, 50);
            System.out.println("*************************************************");
            i += 1000;
        }


    }

}